from airflow import AirflowException
from airflow.models import BaseOperator, DagRun
from airflow.utils.db import provide_session
from airflow.utils.decorators import apply_defaults
from airflow.utils.state import State


class CheckDagRunOperator(BaseOperator):
    """
    Performs a check against the dag run database for given states.
    Raises AirflowException if state is not in given states
    """

    template_fields = ['run_id']
    ui_color = '#fff7e6'

    @apply_defaults
    def __init__(
            self,
            run_id,
            allowed_states=None,
            *args, **kwargs):
        super(CheckDagRunOperator, self).__init__(*args, **kwargs)
        self.run_id = run_id
        self.allowed_states = allowed_states or [State.SUCCESS]
        if not set(self.allowed_states) <= set(State.dag_states):
            raise ValueError(
                'Valid values for `allowed_states`: {}'.format(State.dag_states)
            )

    @provide_session
    def execute(self, context, session=None):
        self.log.info("Checking '%s' for status within %s", self.run_id, self.allowed_states)

        dag_run = DagRun.find(run_id=self.run_id)
        if not dag_run:
            raise AirflowException("Could not find any dag runs with id '{0}'"
                                   .format(self.run_id))
        elif not dag_run.state <= set(self.allowed_states):
            raise AirflowException("Could not find any dag runs with id '{0}' "
                                   "and state within '{1}' but '{2}'"
                                   .format(self.run_id, self.allowed_states, dag_run.state))
        else:
            self.log.info("DAG run with id '%s' checked out ok with state '%s'",
                          self.run_id, dag_run.state)



from unaflow.dags.trigger_dag.trigger_dag_configuration import SENSOR_TASK_ID, TriggerDagConfiguration
from unaflow.operators.gcs_copy import GoogleCloudStorageCopyOperator

from airflow.providers.google.cloud.sensors.gcs import GCSObjectsWithPrefixExistenceSensor
from airflow.sensors.base import BaseSensorOperator
from typing import Sequence, Union
from airflow.models.taskmixin import TaskMixin


XCOM_TOP_FILE = f"{{{{ ti.xcom_pull(task_ids='{SENSOR_TASK_ID}')[0] }}}}"


class GcsMovefilesTriggerDagConfiguration(TriggerDagConfiguration):
    """
    Parameters for Google Cloud Storage, that checks for the existence of objects
     with a prefix.
    This DAG can also be configured to move files before triggering another DAG.
    When triggering another DAG the file(s) that triggered the event are passed to
    the trigger_dag_id, as configuration "filename" or "filenames" if single_file
    is set to False

    :param bucket: The Google cloud storage bucket where the object is.
    :type bucket: str
    :param prefix: The name of the prefix to check in the Google cloud
        storage bucket.
    :type prefix: str
    :param single_file: If destination or destination_bucket is set, we move only one file
    :type single_file: bool
    :param google_cloud_conn_id: The connection ID to use when
        connecting to Google cloud storage.
    :type google_cloud_conn_id: str
    :param destination_bucket:
    :param destination:
    """

    def __init__(self,
                 bucket: str = None,
                 prefix: str = None,
                 destination_bucket=None,
                 destination=None,
                 gcp_conn_id=None,
                 single_file=True,
                 *args, **kwargs
                 ) -> None:
        super(GcsMovefilesTriggerDagConfiguration, self).__init__(*args, **kwargs)
        self.bucket: str = bucket
        self.prefix: str = prefix
        self.destination_bucket = destination_bucket or bucket
        self.destination = destination
        self.gcp_conn_id = gcp_conn_id
        self.configuration = {**self.configuration, 'bucket': self.destination_bucket}
        self.single_file = single_file
        if self.single_file:
            self._source_objects = [XCOM_TOP_FILE]
            self.configuration = {**self.configuration,
                                  'filename': "{{ ti.xcom_pull(task_ids='move_files')[0] }}"}
        else:
            self._source_objects = f"{{{{ ti.xcom_pull(task_ids='{SENSOR_TASK_ID}') }}}}"
            self.configuration = {**self.configuration,
                                  'filenames': "{{ ti.xcom_pull(task_ids='move_files') }}"}

    def create_sensor(self) -> BaseSensorOperator:
        return GCSObjectsWithPrefixExistenceSensor(
            task_id=SENSOR_TASK_ID,
            bucket=self.bucket,
            prefix=self.prefix,
            google_cloud_conn_id=self.gcp_conn_id,
            poke_interval=self.poke_interval,
            mode=self.mode
        )

    def create_downstream_sensor(self) -> Union[TaskMixin, Sequence[TaskMixin]]:
        return GoogleCloudStorageCopyOperator(
            task_id="move_files",
            source_bucket=self.bucket,
            source_objects=self._source_objects,
            destination_bucket=self.destination_bucket,
            destination_object=self.destination,
            move_object=True,
            gcp_conn_id=self.gcp_conn_id,
        )

    def default_doc_md(self):
        return f"""# Triggering DAG for {self.trigger_dag_id}

This DAG listens for files arriving in `{self.bucket}` with the prefix `{self.prefix}`.

Once files arrives we move the file(s) into `{self.destination_bucket}/{self.destination}`
 before we trigger [{self.trigger_dag_id}](/tree?dag_id={self.trigger_dag_id})

The DAG to trigger is supplied the file and information about the triggering dag.

## On failure

Should the DAG fail, the next schedule will start, but it will halt at the first task `{SENSOR_TASK_ID}`,
 since the previous run was a failure.

To fix a failure, you will need to clear the failure, if it is clearable, and then mark the succeeding
DAG run as a failure. This will re-start the DAG run which failed.

If the task that failed is not "clearable", you need to either mark the failed task as succes,
or delete the DAG run, in order to un-clog the DAG."""



from typing import Union

from airflow.models import BaseOperator, SkipMixin, DagRun
from airflow.utils import timezone
from pendulum import datetime


class BranchDagRunOperator(BaseOperator, SkipMixin):
    template_fields = ['execution_dag_id', 'execution_date', ]

    """
    Allows a workflow to "branch" or follow a path following the execution
    of this task.
    """

    def __init__(self,
                 execution_dag_id: str,
                 execution_date: Union[str, datetime],
                 branch_true: str,
                 branch_false: str,
                 *args,
                 **kwargs
                 ):
        super(BranchDagRunOperator, self).__init__(*args, **kwargs)
        self.execution_dag_id = execution_dag_id
        self.execution_date = execution_date
        self.branch_true = branch_true
        self.branch_false = branch_false

    def execute(self, context):
        branch = self.check_dagrun_exists()
        self.skip_all_except(context['ti'], branch)

    def check_dagrun_exists(self):
        if isinstance(self.execution_date, str):
            date = timezone.parse(self.execution_date)
        else:
            date = self.execution_date

        runs = DagRun.find(
            dag_id=self.execution_dag_id,
            execution_date=date,
        )

        if len(runs) > 0:
            return self.branch_true
        else:
            return self.branch_false



import logging
import traceback

import dateutil
from airflow.contrib.hooks.opsgenie_alert_hook import OpsgenieAlertHook

log = logging.getLogger(__name__)


def report_failure_to_opsgenie(message: str = 'Oh no {task_id} in {dag_id} failed',
                               priority: str = 'P3',
                               tags=None,
                               alias=None,
                               responders=None,
                               team_id=None,
                               description='The exception was "{exception}"\nMore: {log_url}',
                               opsgenie_conn_id='opsgenie_default'
                               ):
    """
    Create a function capable of sending an alert to OpsGenie on failure. You can
    for example use this in the on_failure_callback of the DAG as so:
    ...
    on_failure_callback=report_failure_to_opsgenie(message="Houston we have a problem")
    ...
    The params message and description are templated. That means that you can string
    interpolate certain context variables. The variables you can use are:
    task_id: The id of the failed task
    dag_id: The id of the failed DAG
    timestamp: When the failure happened
    date_time: Date representation when the failure happened
    log_url: Airflow's url of the task log
    exception: The exception message
    stacktrace: Stacktrace of the exception
    job_id: The id of the task run
    mark_success_url: Url for directly marking the task as success

    default_args
    :param message: (Templated) Message of the alert
    :param priority: Priority level of the alert.
        Possible values are P1, P2, P3, P4 and P5. Default value is P3.
    :param tags: Tags of the alert.
    :param alias: Client-defined identifier of the alert,
        that is also the key element of Alert De-Duplication.
        See https://docs.opsgenie.com/docs/alert-deduplication
    :param responders: Teams, users, escalations and schedules that the alert
        will be routed to send notifications. type field is mandatory for each
        item, where possible values are team, user, escalation and schedule.
        If the API Key belongs to a team integration, this field will be overwritten
        with the owner team. Either id or name of each responder should be provided.
        You can refer below for example values.
    :param team_id: If this is set, it will create one responder with this id
    :param description: (Templated) Description field of the alert that is generally used to provide
        a detailed information about the alert.
    :param opsgenie_conn_id: The name of the Opsgenie connection to use
    :return: A function that can execute on failure context, and which itself returns
        OpsgenieAlertHook.execute result
    """

    def on_failure(context):
        logging.info("Hooking into OpsGenie")
        context_map = {'dag_id': context.get('task_instance').dag_id,
                       'task_id': context.get('task_instance').task_id,
                       'timestamp': context.get('ts'),  # Timestamp Example: 2018-01-01T00:00:00+00:00
                       'log_url': context.get('task_instance').log_url,
                       'date_time': dateutil.parser.parse(context.get('ts')),
                       'exception': "%s" % str(context['exception']).replace('"', '').replace("'", ""),
                       'stacktrace': '%s' % traceback.format_exc().replace('"', '').replace("'", ""),
                       'job_id': "%s" % context['task_instance'].job_id,
                       'mark_success_url': context.get('task_instance').mark_success_url
                       }
        _responders = [{"id": team_id, "type": "team"}] if team_id else responders
        json = {
            "message": message.format(**context_map),
            "description": description.format(**context_map),
            "responders": _responders,
            "tags": tags,
            "details": {
                "logUrl": context_map['log_url'],
                "stacktrace": context_map['stacktrace'],
                "jobId": context_map['job_id'],
                "dagId": context_map['dag_id'],
                "taskId": context_map['task_id']
            },
            "priority": priority,
            "alias": alias
        }

        logging.info("Trying to send OpsGenie this json: %s", json)

        hook = OpsgenieAlertHook(opsgenie_conn_id)
        return hook.execute(json)

    return on_failure



from typing import Union

from airflow import DAG
from airflow.models import BaseOperator, DagBag
from airflow.utils import timezone
from airflow.utils.decorators import apply_defaults
from pendulum import datetime


class ClearDagOperator(BaseOperator):

    template_fields = ['execution_dag_id', 'execution_date']

    @apply_defaults
    def __init__(
            self,
            execution_dag_id: str,
            execution_date: Union[str, datetime],
            *args,
            **kwargs
    ):
        super(ClearDagOperator, self).__init__(*args, **kwargs)
        self.execution_dag_id = execution_dag_id
        self.execution_date = execution_date

    def execute(self, context):
        self.clear_dag()

    def clear_dag(self):
        if isinstance(self.execution_date, str):
            date = timezone.parse(self.execution_date)
        else:
            date = self.execution_date

        clear_dag: DAG = DagBag().get_dag(self.execution_dag_id)
        clear_dag.clear(
            start_date=date,
            end_date=date
        )



from airflow.models import DagRun
from airflow.sensors.base_sensor_operator import BaseSensorOperator
from airflow.utils.db import provide_session
from airflow.utils.decorators import apply_defaults
from airflow.utils.state import State


class ExternalDagRunSensor(BaseSensorOperator):
    """
    Waits for a different DAG or a task in a different DAG to complete for a
    specific execution_date

    :param external_run_id: The run_id of the dag you wait for
    :type external_run_id: str
    :param allowed_states: list of allowed states, default is ``['success', 'failed']``
    :type allowed_states: list
    """
    template_fields = ['external_run_id']
    ui_color = '#19647e'

    @apply_defaults
    def __init__(self,
                 external_run_id,
                 allowed_states=None,
                 *args,
                 **kwargs):
        super(ExternalDagRunSensor, self).__init__(*args, **kwargs)
        self.allowed_states = allowed_states or [State.SUCCESS, State.FAILED]
        if not set(self.allowed_states) <= set(State.dag_states):
            raise ValueError(
                'Valid values for `allowed_states`: %s ' % State.dag_states
            )
        self.external_run_id = external_run_id

    @provide_session
    def poke(self, context, session=None):
        self.log.info('Poking for %s... ', self.external_run_id)
        DR = DagRun

        count = session.query(DR).filter(
            DR.run_id == self.external_run_id,
            DR.state.in_(self.allowed_states),
        ).count()

        session.commit()
        return count == 1



import ast
from os.path import basename

from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults


class GoogleCloudStorageCopyOperator(BaseOperator):
    """
    Copies objects from a bucket to another, with renaming if requested.

    :param source_bucket: The source Google cloud storage bucket where the
         object is. (templated)
    :type source_bucket: str
    :param source_objects: The source names of the objects to copy in the Google cloud
        storage bucket. (templated)
    :type source_object: str
    :param destination_bucket: The destination Google cloud storage bucket
        where the object should be. If the destination_bucket is None, it defaults
        to source_bucket. (templated)
    :type destination_bucket: str
    :param destination_object: If supplied, the basename of the source-object will
        will be appended to this object. (templated)
    :type destination_object: str
    :param move_object: When move object is True, the object is moved instead
        of copied to the new location. This is the equivalent of a mv command
        as opposed to a cp command.
    :type move_object: bool
    :param google_cloud_storage_conn_id: The connection ID to use when
        connecting to Google cloud storage.
    :type google_cloud_storage_conn_id: str
    :param delegate_to: The account to impersonate, if any.
        For this to work, the service account making the request must have
        domain-wide delegation enabled.
    :type delegate_to: str
    :param last_modified_time: When specified, the objects will be copied or moved,
        only if they were modified after last_modified_time.
        If tzinfo has not been set, UTC will be assumed.
    :type last_modified_time: datetime.datetime


    """
    template_fields = ('source_bucket', 'source_objects', 'destination_bucket',
                       'destination_object',)
    ui_color = '#f0eee4'

    @apply_defaults
    def __init__(self,
                 source_bucket,
                 source_objects,
                 destination_bucket=None,
                 destination_object=None,
                 move_object=False,
                 gcp_conn_id='google_cloud_default',
                 delegate_to=None,
                 last_modified_time=None,
                 *args,
                 **kwargs):
        super(GoogleCloudStorageCopyOperator,
              self).__init__(*args, **kwargs)
        self.source_bucket = source_bucket
        self.source_objects = source_objects
        self.move_object = move_object
        self.destination_bucket = destination_bucket
        self.destination_object = destination_object
        self.gcp_conn_id = gcp_conn_id
        self.delegate_to = delegate_to
        self.last_modified_time = last_modified_time

    def execute(self, context):

        hook = GCSHook(
            gcp_conn_id=self.gcp_conn_id,
            delegate_to=self.delegate_to
        )

        if self.destination_bucket is None:
            self.log.warning(
                'destination_bucket is None. Defaulting it to source_bucket (%s)',
                self.source_bucket)
            self.destination_bucket = self.source_bucket

        source_objects = ast.literal_eval(self.source_objects) if isinstance(self.source_objects,
                                                                             str) else self.source_objects
        files_copied = []
        for source_object in source_objects:
            if self.destination_object is None:
                destination_object = source_object
            else:
                separator = "" if self.destination_object[-1:] == "/" else "/"
                destination_object = "%s%s%s" % (self.destination_object,
                                                 separator,
                                                 basename(source_object))

            self._copy_single_object(hook=hook, source_object=source_object,
                                     destination_object=destination_object)
            files_copied.append(destination_object)

        return files_copied

    def _copy_single_object(self, hook, source_object, destination_object):
        if self.last_modified_time is not None:
            # Check to see if object was modified after last_modified_time
            if hook.is_updated_after(self.source_bucket,
                                     source_object,
                                     self.last_modified_time):
                self.log.debug("Object has been modified after %s ", self.last_modified_time)
                pass
            else:
                return

        self.log.info('Executing copy of gs://%s/%s to gs://%s/%s',
                      self.source_bucket, source_object,
                      self.destination_bucket, destination_object)

        hook.rewrite(self.source_bucket, source_object,
                     self.destination_bucket, destination_object)

        if self.move_object:
            hook.delete(self.source_bucket, source_object)
